#!/usr/bin/wish
# Authors: Claudio Esperança and Paulo Roma.
#
proc msec { hh mm ss mmss } {
   return [expr  $mmss+($ss+($mm+$hh*60)*60)*1000]
}

proc msecfromstring { s } {
   if {[scan $s "%d:%d:%d,%d" h1 m1 s1 ms1]!=4} { 
      if {[scan $s "%d:%d:%d.%d" h1 m1 s1 ms1]!=4} { 
         error "Ill formed time stamp: $s"
      } else {
	   return [msec $h1 $m1 $s1 $ms1]
      }
   } else {
      return [msec $h1 $m1 $s1 $ms1]
   }
}

proc msecfromstring2 { s } {
   if {[scan $s "%d:%d:%d.%d" h1 m1 s1 ms1]!=4} { 
      error "Ill formed time stamp: $s"
   } else {
      return [msec $h1 $m1 $s1 [expr $ms1*10]]
   }
}

proc time { mmss } {
   if {$mmss<0} { 
      set mmss [expr -$mmss]
      set sign "-"
   } else {
      set sign ""
   }
   set ms [expr $mmss % 1000]
   set mmss [expr $mmss / 1000]
   set ss [expr $mmss % 60]
   set mmss [expr $mmss /60]
   set mm [expr $mmss % 60]
   set mmss [expr $mmss /60]
   set hh $mmss
   return [format "$sign%02d:%02d:%02d,%03d" $hh $mm $ss $ms]
}

proc fpsdialog {} {
   set answer [tk_dialog .d "FPS" "How many frames per second?" questhead 0 \
      "23.976" "25" "29.970" "30" "Cancel"]
   if {$answer==0} {return 23.976}
   if {$answer==1} {return 25}
   if {$answer==2} {return 29.970}
   if {$answer==3} {return 30}
   return 0
}

proc loadsubs {} {
   global subarray
   .subs.lb delete 0 end
   foreach i [lsort -integer [array names subarray]] { 
      set from [lindex $subarray($i) 0]
      set to [lindex $subarray($i) 1]
      set sub [lindex $subarray($i) 2]
      set sub [join [split $sub "\n"] "|"]
      .subs.lb insert end "[time $from] $sub"
   }
}

proc loadtofrom {} {
   global subarray sub1index sub2index sub1src sub1dst sub2src sub2dst
   set timestamp [time [lindex $subarray($sub2index) 0]]
   set sub2src $timestamp
   set sub2dst $timestamp
   set timestamp [time [lindex $subarray($sub1index) 0]]
   set sub1src $timestamp
   set sub1dst $timestamp
}

proc doopen {} {
   global dir
   set types {
                  {{sub Files}       {.sub .srt}    }
                  {{All Files}        *             }
              }
   set filename [tk_getOpenFile -initialdir $dir -filetypes $types]   
   if {$filename==""} return
   openfile $filename
}

proc openfile { filename } {
   global dir
   set dir [file dirname $filename]
   set f [open $filename r]
   while {![eof $f]} {
      gets $f line
      if {[string length $line]==0} continue
      if {[string first \[ $line]==0} continue
      break
   }
   close $f
   if {[scan $line "\{%d\}\{%d\}%\[^~\]" a b c]==3} {
      doopensub $filename
   } elseif {[scan $line "%d:%d:%d.%d" h1 m1 s1 ms1]==4} {
      doopensrt2 $filename
   } elseif {[scan $line "%d" i]==1} {
      doopensrt $filename
   } else {
      error "Unsupported subtitle format:\n$line"
   }
   global sub1index sub2index subarray
   set sub1index 1
   set sub2index [expr [array size subarray]]
   loadtofrom
}

proc doopensub { filename } {
   global subarray sub1src sub1dst sub2src sub2dst msg 
   set fps [fpsdialog]
   if {$fps==0} return
   catch "unset subarray"
   set f [open $filename r]
   set i 0
   while {![eof $f]} {
      gets $f line
      set n [scan $line "\{%d\}\{%d\}%\[^~\]" a b c]
      if {$n!=3} continue
      incr i
      set from [expr int($a * 1.0 / $fps * 1000)]
      set to [expr int($b * 1.0 / $fps * 1000)]
      set sub [join [split $c "|"] "\n"]
      set subarray($i) [list $from $to $sub] 
      set timestamp [time $from]
      if {$i==1} {
      	 set sub1src $timestamp
      	 set sub1dst $timestamp
      } else {
         set sub2src $timestamp
	 set sub2dst $timestamp
      }      
   }
   loadsubs
   set msg "$filename loaded with $i subtitles"
   tk_messageBox -message $msg
}

proc doopensrt {filename} {
   global subarray sub1src sub1dst sub2src sub2dst msg
   catch "unset subarray"
   set f [open $filename r]
   set i 0
   set blankline 0
   while {![eof $f]} {
      gets $f line
#      puts $line
      if {$line==""} { incr blankline ; continue }
      set j [lindex $line 0]
      #if {$j!=$i+1} { puts "Erro na sequencia : $j" ; exit }
      incr i
      gets $f line
      set timestamp1 [lindex $line 0]
      set timestamp2 [lindex $line 2]
      set from [msecfromstring $timestamp1]
      set to [msecfromstring $timestamp2]
      if {$i==1} {
      	 set sub1src $timestamp1
      	 set sub1dst $timestamp1
      } else {
         set sub2src $timestamp1
	 set sub2dst $timestamp1
      }
      gets $f line
      set sub $line
      gets $f line
      while {$line!=""} {
	 append sub "\n" $line
	 gets $f line
      }
      set subarray($i) [list $from $to $sub] 
   }
   close $f
   loadsubs
   set msg "$filename loaded with $j subtitles"
}

proc doopensrt2 {filename} {
   global subarray sub1src sub1dst sub2src sub2dst msg
   catch "unset subarray"
   set f [open $filename r]
   set j 0
   set blankline 0
   while {![eof $f]} {
      gets $f line
      if {$line==""} { incr blankline ; continue }	
      if {[string first \[ $line]==0} continue
      incr j
      set ts [split $line ","]
      set timestamp1 [lindex $ts 0]
      set timestamp2 [lindex $ts 1]
      set from [msecfromstring2 $timestamp1]
      set to [msecfromstring2 $timestamp2]
      if {$j==1} {
      	 set sub1src $timestamp1
      	 set sub1dst $timestamp1
      } else {
         set sub2src $timestamp1
	 set sub2dst $timestamp1
      }
      gets $f line
      while {[set i [string first {[br]} $line]]>0} {
      	 set line [string replace $line $i [expr $i+3] "\n"]
      }
      set subarray($j) [list $from $to $line] 
      gets $f line
   }
   close $f
   loadsubs
   set msg "$filename loaded with $j subtitles"
}

proc computefactors { {shiftonly 0} } {
   global delay scale sub1src sub1dst sub2src sub2dst
   set x0 [msecfromstring $sub1src]
   set y0 [msecfromstring $sub1dst]
   set x1 [msecfromstring $sub2src]
   set y1 [msecfromstring $sub2dst]
   if {$shiftonly} {
      set scale 1.0 
      set delay [expr $y0 - $x0]
   } else {
      set scale [expr 1.0 * ($y1 - $y0) / ($x1 - $x0)]
      set delay [expr $y0 - 1.0 * $scale * $x0]   
   }
}

proc subtitlerange { {question "Do op on what subtitle range?"} } {
   global subarray
   set sel [.subs.lb curselection]
   if {[llength $sel]>1} {
      set answer [tk_dialog .d "Subtitle Range" $question questhead 0 \
         "All subtitle lines" "Only selected lines"]
   } else {
      set answer 0
   }
   if {$answer == 0} {
      set result [lsort -integer [array names subarray]]
   } else {
      foreach s $sel { lappend result [expr $s+1] }
   }
   return $result
}

proc applyfactors {} {
   global subarray delay scale msg
   foreach i [subtitlerange] { 
      set from [lindex $subarray($i) 0]
      set to [lindex $subarray($i) 1]
      set sub [lindex $subarray($i) 2]      
      set newfrom [expr int($from * $scale + $delay)]
      set newto [expr int ($to * $scale + $delay)]    
      set subarray($i) [list $newfrom $newto $sub]
   }
   loadsubs
   loadtofrom
   set msg "Subs Resynched"
}

proc dosavesrt {} {
   global subarray delay scale msg dir
   set types {
                  {{srt Files}       {.srt}        }
                  {{All Files}        *             }
              }
   set filename [tk_getSaveFile -initialdir $dir -filetypes $types]   
   if {$filename==""} return
   set f [open $filename w]
   fconfigure $f -translation crlf
   set j 0
   set lastfrom 0
   foreach i [subtitlerange "What subtitle range is to be saved?"] { 
      set from [lindex $subarray($i) 0]
      set to [lindex $subarray($i) 1]
      set sub [lindex $subarray($i) 2]   
      if {$from<$lastfrom} {
      	 # Do not allow time to go back   
      	 continue 
      }
      puts $f [incr j]
      puts $f "[time $from] --> [time $to]"
      puts $f "$sub\n"
      set lastfrom $from
   }
   close $f
   set msg "$filename saved"
}

proc doselfirst {} {
   global sub1src sub1dst sub2src sub2dst subarray sub1index 
   set i [expr [.subs.lb index active]+1]
   set timestamp [time [lindex $subarray($i) 0]]
   set sub1index $i
   set sub1src $timestamp
   set sub1dst $timestamp
}

proc dosellast {} {
   global sub1src sub1dst sub2src sub2dst subarray sub2index
   set i [expr [.subs.lb index active]+1]
   set timestamp [time [lindex $subarray($i) 0]]
   set sub2index $i
   set sub2src $timestamp
   set sub2dst $timestamp
}

proc dooffset {} {
   computefactors 1
   applyfactors 
}

proc dooffsetscale {} {
   computefactors 0
   applyfactors 
}

set sub1src ""
set sub1dst ""
set sub2src ""
set sub2dst ""
set dir "." 

pack [frame .f] -fill x
pack [button .f.opensub -text "open" -command doopen] -side left
pack [button .f.savesrt -text "save" -command dosavesrt] -side left
pack [button .f.quit -text "quit" -command exit] -side right
pack [frame .g] -fill x
pack [label .g.sub1label1 -text "1st sub"] \
     [entry .g.sub1src -textvar sub1src] \
     [label .g.sub1label2 -text " => "] \
     [entry .g.sub1dst -textvar sub1dst] \
     [button .g.offset -text "apply offset only" -command dooffset] -side left
pack [frame .h] -fill x
pack [label .h.sub2label1 -text "2nd sub"] \
     [entry .h.sub2src -textvar sub2src] \
     [label .h.sub2label2 -text " => "] \
     [entry .h.sub2dst -textvar sub2dst] \
     [button .h.offsetscale -text "apply offset & scale" -command dooffsetscale] \
     -side left
pack [frame .subs] [label .msg -textvar msg] -side bottom -fill both
pack [frame .subs.sel] -side bottom
pack [button .subs.sel.first -text "Sel 1st" -command doselfirst] \
     [button .subs.sel.last -text "Sel 2nd" -command dosellast] \
     -side left    
pack [listbox .subs.lb -exportselection 0 -selectmode extended \
     -yscroll {.subs.sb set}] -side left -expand yes -fill both
pack [scrollbar .subs.sb -command {.subs.lb yview}] -side right -fill y


if {$argc>0} { openfile [lindex $argv end] }

